"""Implementation of GBA scaled flux Problem class.

Based on Hugo Dourado scaled flux GBA optimization

Ref: Presentation
Growth Mechanics: the Economy, Control and Optimality of Self-Replication
Metabolic Pathway Analysis 2021, Hugo Dourado

First step: convert XBA into Hugo RGBA format
Second step: implement optimization (inline with Hugo formalism)

Peter Schubert, HHU Duesseldorf, September 2022
"""

import numpy as np
import pandas as pd
import re
import sympy as sp
import sbmlxdf

from xbanalysis.model.xba_unit_def import XbaUnit
from xbanalysis.utils.utils import expand_kineticlaw


class GbaSfluxProblem:

    def __init__(self, xba_model):
        """Initialize GbaSfluxProblem from a xba_model.

        :param xba_model: xba model with all required GBA parameters
        :type xba_model: XbaModel
        """
        self.xba_model = xba_model

        # while XBA model contains enzymes for all metabolic reactions with related synthesis reactions,
        #  in scaled flux GBA optimization one a single protein (here we use the ribosome) is used
        metab_sids = [s.id for s in xba_model.species.values() if s.sboterm.is_in_branch('SBO:0000247')]
        self.const_metab_sids = [sid for sid in metab_sids if xba_model.species[sid].constant is True]
        self.var_metab_sids = [sid for sid in metab_sids if xba_model.species[sid].constant is False]
        sid_ribo = [s.id for s in xba_model.species.values() if s.sboterm.is_in_branch('SBO:0000250')][0]
        self.var_sids = self.var_metab_sids + [sid_ribo]
        self.sids = self.const_metab_sids + self.var_sids
        self.sid2idx = {msid: idx for idx, msid in enumerate(self.sids)}

        self.mr_rids = [r.id for r in xba_model.reactions.values()
                        if r.sboterm.is_in_branch('SBO:0000167')]
        rid_ps_ribo = xba_model.species[sid_ribo].ps_rid
        self.rids = self.mr_rids + [rid_ps_ribo]
        self.rid2idx = {mrid: idx for idx, mrid in enumerate(self.rids)}

        self.mw_sids = np.array([xba_model.species[sid].mw for sid in self.sids])
        self.mw_enz = np.array([xba_model.species[xba_model.reactions[rid].enzyme].mw
                                for rid in self.rids])
        # mw_enz is used for scaling of kcats, weight for protein synthesis needs to be rescaled
        s = xba_model.species[sid_ribo]
        if hasattr(s, 'rba_process_costs'):
            p_len = s.rba_process_costs['translation']
        else:
            r = xba_model.reactions[rid_ps_ribo]
            p_len = min(r.reactants.values())
        self.mw_enz[-1] *= p_len

        # retrieve S-matrix: internal metabos + total protein and metabolic reaction + ribosome synthesis
        sids_x_rids = xba_model.get_stoic_matrix(self.sids, self.rids)

        # calulate mass turnover in gram reactants/products per mol reactions (g/mol)
        sids_x_rids_react = np.where(sids_x_rids > 0, 0, sids_x_rids)
        sids_x_rids_prod = np.where(sids_x_rids < 0, 0, sids_x_rids)
        mass_turnover_react = -sids_x_rids_react.T.dot(self.mw_sids)
        mass_turnover_prod = sids_x_rids_prod.T.dot(self.mw_sids)
        self.mass_turnover = np.max((mass_turnover_react, mass_turnover_prod), axis=0)

        # mass normalized stoichiometric matrix
        self.sidx_x_rids_mn = (sids_x_rids * self.mw_sids.reshape(-1, 1)) / self.mass_turnover

        self.rho = xba_model.parameters['rho'].value
        self.ext_conc_mn = np.array([xba_model.species[sid].initial_conc * xba_model.species[sid].mw
                                     for sid in self.const_metab_sids])
        self.var_initial_mn = np.array([xba_model.species[sid].initial_conc * xba_model.species[sid].mw
                                        for sid in self.var_metab_sids])

        # identify unit ids for kinetic parameters
        units_per_s = [XbaUnit({'kind': 'second', 'exp': -1.0, 'scale': 0, 'mult': 1.0})]
        units_mol_per_l = [XbaUnit({'kind': 'litre', 'exp': -1.0, 'scale': 0, 'mult': 1.0}),
                           XbaUnit({'kind': 'mole', 'exp': 1, 'scale': 0, 'mult': 1.0})]
        self.kcat_udid = self.get_unit_def_id(units_per_s)
        self.km_udid = self.get_unit_def_id(units_mol_per_l)
        self.kcats_global = {pid: p.value for pid, p in xba_model.parameters.items()
                             if p.units == self.kcat_udid}
        self.kms_global = {pid: p.value for pid, p in xba_model.parameters.items()
                           if p.units == self.km_udid}

    def get_unit_def_id(self, query_units):
        for udid, ud in self.xba_model.unit_defs.items():
            if ud.is_equivalent(query_units):
                return udid
        return None

    def get_kcats(self, rid):

        r = self.xba_model.reactions[rid]
        kcats_local = {pid: val for pid, [val, units] in r.local_params.items() if units == self.kcat_udid}

        # analyse kinetic law to identify kcat_f and kcat_r
        expanded_kl = sbmlxdf.misc.mathml2numpy(r.kinetic_law)
        kl = expand_kineticlaw(expanded_kl, self.xba_model.functions)
        used_kcats = {kcat_id: val for kcat_id, val in (self.kcats_global | kcats_local).items()
                      if re.search(r'\b' + kcat_id + r'\b', kl)}
        kcat_f = 0.0
        kcat_r = 0.0
        for kcat_id, val in used_kcats.items():
            # replace other parameters with value of 1.0 (exclude selected kcat and numbers)
            kl_kcat = re.sub(r'\b(?!(' + kcat_id + r'\b))[a-zA-Z_]\w*\b', str(1.0), kl)
            # set two different kcat values to check which increases/reduces the rate
            kl_kcat_v2 = re.sub(r'\b' + kcat_id + r'\b', str(2.0), kl_kcat)
            kl_kcat_v10 = re.sub(r'\b' + kcat_id + r'\b', str(10.0), kl_kcat)
            if sp.sympify(kl_kcat_v2) > sp.sympify(kl_kcat_v10):
                kcat_r = val
            else:
                kcat_f = val
        return kcat_f, kcat_r

    def get_kms(self, rid):
        """

        :param rid: reaction id
        :type rid: string
        :return:
        :rtype:
        """

        r = self.xba_model.reactions[rid]
        kms_local = {pid: val for pid, [val, units] in r.local_params.items() if units == self.km_udid}

        # analyse kinetic law to identify Km and Ki values
        expanded_kl = sbmlxdf.misc.mathml2numpy(r.kinetic_law)
        kl = expand_kineticlaw(expanded_kl, self.xba_model.functions)
        used_kms = {km_id: val for km_id, val in (self.kms_global | kms_local).items()
                    if re.search(r'\b' + km_id + r'\b', kl)}
        kms = {}
        kis = {}
        # check correlation between metabolite and KM value, scaling both equally should not change value of kl
        for km_id, val in used_kms.items():
            for sid in r.reactants:
                # in first step: replace parameters with value of 1.0 (excluding selected km and sid)
                #   km_id, sid are excluded from matching, e.g. \b(?!(' + km_id + r'\b))
                #   parameter names starting with letter or underscore [a-zA-Z_] are matched
                #   numbers are not matched, as they do not start with [a-zA-Z_]
                # in second phase: protected km_id and sid are replaced by values
                kl_km_m = re.sub(r'\b(?!(' + km_id + r'\b))(?!(' + sid + r'\b))[a-zA-Z_]\w*\b', str(1.0), kl)
                kl_km_m_v2 = re.sub(r'\b[a-zA-Z_]\w*\b', str(2.0), kl_km_m)
                kl_km_m_v10 = re.sub(r'\b[a-zA-Z_]\w*\b', str(10.0), kl_km_m)
                if bool(abs(sp.sympify(kl_km_m_v2) - sp.sympify(kl_km_m_v10)) <= 1e-8):
                    kms[sid] = val
            for sid in r.products:
                kl_km_m = re.sub(r'\b(?!(' + km_id + r'\b))(?!(' + sid + r'\b))[a-zA-Z_]\w*\b', str(1.0), kl)
                kl_km_m_v2 = re.sub(r'\b[a-zA-Z_]\w*\b', str(2.0), kl_km_m)
                kl_km_m_v10 = re.sub(r'\b[a-zA-Z_]\w*\b', str(10.0), kl_km_m)
                if abs(sp.sympify(kl_km_m_v2) - sp.sympify(kl_km_m_v10)) <= 1e-8:
                    kis[sid] = val
        return kms, kis

    def mass_normalized_model(self):
        """Retrieve mass normalized model

        :return: gba model mass normalized as per Hugo Dourado
        :rtype: dict of pandas DataFrames
        """
        protein_mn = self.rho - sum(self.var_initial_mn)

        kcats = np.array([self.get_kcats(rid) for rid in self.rids]).T
        skcats = kcats * 3600.0 * self.mass_turnover / self.mw_enz

        km = np.zeros_like(self.sidx_x_rids_mn)
        ki = np.zeros_like(self.sidx_x_rids_mn)
        for rid in self.rids:
            kms, kis = self.get_kms(rid)
            for sid, val in kms.items():
                km[self.sid2idx[sid], self.rid2idx[rid]] = val
            for sid, val in kis.items():
                ki[self.sid2idx[sid], self.rid2idx[rid]] = val
        # scale Michaelis constants with molecluar weights
        skm = km * self.mw_sids.reshape(-1, 1)
        ski = ki * self.mw_sids.reshape(-1, 1)
        ska = np.zeros_like(self.sidx_x_rids_mn)

        mn_model = {
            'N': pd.DataFrame(self.sidx_x_rids_mn, index=self.sids, columns=self.rids),
            'KM': pd.DataFrame(skm, index=self.sids, columns=self.rids),
            'KI': pd.DataFrame(ski, index=self.sids, columns=self.rids),
            'KA': pd.DataFrame(ska, index=self.sids, columns=self.rids),
            'kcat': pd.DataFrame(skcats, index=['kcat_f', 'kcat_b'], columns=self.rids),
            'conditions': pd.DataFrame(np.hstack((self.rho, self.ext_conc_mn)),
                                       index=['rho'] + self.const_metab_sids, columns=[1]),
            'lower_c': pd.DataFrame(np.zeros(len(self.var_sids)),
                                    index=self.var_sids, columns=['lower']),
            'initial': pd.DataFrame(np.hstack((self.var_initial_mn, protein_mn)),
                                    index=self.var_sids, columns=['initial']),
        }
        return mn_model
